"""Utilities for launching processes in background tasks."""
import os
import subprocess
from multiprocessing import Process


def detachify(func):
    """Decorate a function so that its calls are async in a detached process.
    
    Usage
    -----

    .. code::
        import time

        @detachify
        def f(message):
            time.sleep(5)
            print(message)

        f('Async and detached!!!')

    """
    # create a process fork and run the function
    def forkify(*args, **kwargs):
        if os.fork() != 0:
            return
        func(*args, **kwargs)

    # wrapper to run the forkified function
    def wrapper(*args, **kwargs):
        proc = Process(target=lambda: forkify(*args, **kwargs))
        proc.start()
        proc.join()
        return

    return wrapper


def shell_command(command_str):
    """Run a shell command as a string, returning a tuple of (stdout, stderr)."""
    try:
        out = subprocess.check_output(command_str, shell=True, stderr=subprocess.STDOUT)
        if out is not None:
            out = out.decode()
        err = None
    except subprocess.CalledProcessError as e:
        err = e.output.decode()
        out = None

    return out, err



"""Tests for the cli module."""
import subprocess

import pytest
from click.testing import CliRunner

from shlack import slack
from shlack.cli import message as message_cli
from shlack.cli import task as task_cli


class MockChat:
    """Mock chat class for slacker."""

    @staticmethod
    def post_message(*args, **kwargs):
        """Mock the post message method."""
        return None


@pytest.mark.parametrize(
    "command, is_valid",
    [
        ([], True),
        (["echo", "1"], True),
        (["INVALD_COMMAND"], False),
        (["echo", "1", "-f", "text"], True),
        (["echo", "1", "-f", "auto"], True),
        (["echo", "1", "-f", "file"], True),
        (["echo", "1", "-f", "none"], True),
        (["echo", "1", "-f", "invalid"], False),
    ],
)
def test_task(monkeypatch, command, is_valid):
    """Test that the task logic does not except out.
    
    This is NOT a test of the slack message content. I have no idea how to test that.
    """
    monkeypatch.setattr("slacker.Chat", lambda *args, **kw: MockChat())
    monkeypatch.setattr(
        "shlack.slack.upload_file_get_permalink", lambda *a, **kw: "URL"
    )
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")
    monkeypatch.setenv("SLACK_CHANNEL", "ALSO_FAKE")

    runner = CliRunner()
    result = runner.invoke(task_cli.main, command + ["--no-detach"])

    if is_valid:
        assert result.exit_code == 0
    else:
        assert result.exit_code != 0


@pytest.mark.parametrize(
    "args",
    [
        tuple(),
        ("OK",),
        ("OK", "-a", "1", "2"),
        ("OK", "-a", "1", "2", "-a", "3", "4"),
        ("OK", "-u", "readme.md"),
        ("OK", "-a", "1", "2", "-u", "readme.md"),
        ("OK", "-a", "1", "2", "-a", "3", "4", "-u", "readme.md"),
        ("-a", "1", "2"),
        ("-a", "1", "2", "-a", "3", "4"),
        ("-u", "readme.md"),
        ("-a", "1", "2", "-u", "readme.md"),
        ("-a", "1", "2", "-a", "3", "4", "-u", "readme.md"),
    ],
)
def test_message(monkeypatch, args):
    """Test that the message logic does not except out.
    
    This is NOT a test of the slack message content. I have no idea how to test that.
    """
    monkeypatch.setattr("slacker.Chat", lambda *args, **kw: MockChat())
    monkeypatch.setattr(
        "shlack.slack.upload_file_get_permalink", lambda *a, **kw: "URL"
    )
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")
    monkeypatch.setenv("SLACK_CHANNEL", "ALSO_FAKE")

    runner = CliRunner()
    result = runner.invoke(message_cli.main, args)

    assert result.exit_code == 0


@pytest.mark.parametrize(
    "args",
    [
        ("shlack", "--help"),
        ("shlack", "task", "--help"),
        ("shlack", "message", "--help"),
        ("python", "-m", "shlack", "--help"),
        ("python", "-m", "shlack", "task", "--help"),
        ("python", "-m", "shlack", "message", "--help"),
        ("python", "-m", "shlack.cli", "--help"),
        ("python", "-m", "shlack.cli", "task", "--help"),
        ("python", "-m", "shlack.cli", "message", "--help"),
        ("python", "-m", "shlack.cli.task", "--help"),
        ("python", "-m", "shlack.cli.message", "--help"),
    ],
)
def test_cli_installed(args):
    """Test that the CLI has been installed."""
    subprocess.check_output(args)



"""CLI handler for running a task."""
import sys
from io import StringIO

import click

from shlack import runner, slack

from . import common_options

_detach_help = (
    "Option to run the task in a detached process. "
    + "This is useful when you want to close your terminal / ssh session and let a "
    + "long task run. Default --detach."
)
_format_help = (
    "Format of stdout and stderr messages to send. If 'attach', the data will be "
    + "sent as an attachment. If 'file' the data will be uploaded as a file. If "
    + "'auto', shlack will choose between file and text depending on the content "
    + "length. If 'none', nothing will be sent. Default 'auto'."
)


@click.command()
@click.argument("task", nargs=-1, type=str)
@click.option("--detach/--no-detach", "detach", default=True, help=_detach_help)
@click.option(
    "-f",
    "--out-format",
    "out_format",
    type=click.Choice({"text", "file", "auto", "none"}),
    default="auto",
    help=_format_help,
)
@common_options
def main(oauth_api_token, channel, detach, out_format, task):
    """Run a task and send a slack message after.
    
    $ shlack task 'sleep 5 && echo "FROM THE PAST!"' -c ... -t ...

    Wait 5 seconds and you should see a nicely formatted message in your slack 
    workspace.

    By default the task runs in a detached process and you'll only know its complete 
    when the slack message comes in. You can 

    """
    # if task is not defined, we are done
    if not task:
        click.echo("shlack task called without a task!")
        sys.exit(0)

    # define messenger
    slacker = slack.slacker_factory(api_key=oauth_api_token)
    command_joined = " ".join(task)

    def f():
        out, err = runner.shell_command(command_joined)

        # always attach a copy of the command
        attachments = [
            slack.attachment_formatter(
                {"": "```$ %s```" % (command_joined)}, color="#000000"
            )
        ]

        # add more attachments if asked for
        if out and out_format != "none":
            attachments.append(
                slack.attach_text_data(slacker, out, "stdout", out_format, "#439FE0")
            )

        if err and out_format != "none":
            attachments.append(
                slack.attach_text_data(slacker, err, "stderr", out_format, "danger")
            )

        # send the data
        slacker.chat.post_message(channel, "", attachments=attachments)

        # echo out if not detached
        if not detach:
            if out:
                click.echo(out)
            if err:
                click.echo(err)

        return 0 if err is None else 1

    if detach:
        runner.detachify(f)()
        sys.exit(0)

    # otherwise run in non-detached and exit if error
    sys.exit(f())


if __name__ == "__main__":
    main()



"""Tests for the runner module."""
import os
import time

from shlack.runner import detachify, shell_command


def test_detachify(tmp_path):
    """Test that deatchification does not result in an error.
    
    I don't know how to test os.fork but have verified it has the expected result 
    for the vanilla case. It checks

    - That the process was detached
    - And a known string was written to a file
    """
    now_str = str(round(time.time(), 3))
    tmp_file = tmp_path / "temporary.txt"

    def write_string(s):
        """Write a string to the temp file."""
        time.sleep(0.2)
        with open(str(tmp_file), "w") as f:
            f.write(s)

    # run it and assert that it took less than the wait time to move on
    # if it took _longer_ than the function might not have been detached.
    detached_fun = detachify(write_string)
    t1 = time.time()
    detached_fun(now_str)
    t2 = time.time()
    assert (t2 - t1) < 0.2

    # wait a little and check the file that was written.
    time.sleep(0.3)

    with open(str(tmp_file), "r") as f:
        assert now_str == f.read()


def test_detachify_env_persist(monkeypatch, tmp_path):
    """Test that the user env is used in the deatched process."""
    tmp_file = tmp_path / "temporary.txt"
    now_str = str(round(time.time(), 3))
    monkeypatch.setenv("ENV_VARIABLE", now_str)

    def f():
        os.system("echo $ENV_VARIABLE > %s" % tmp_file)

    detachify(f)()
    time.sleep(0.05)

    with open(str(tmp_file), "r") as f:
        assert now_str == f.read().strip()


def test_shell_good_command():
    """Test that shell command works when the command is known to work."""
    out, err = shell_command("ls .")
    assert err is None
    assert "test" in out


def test_shell_bad_command():
    """Test that shell command works when the command is known to work."""
    out, err = shell_command("ls adasdasdas")
    assert out is None
    assert "adasdasdas" in err



"""CLI handler for running a task."""
import sys

import click

from shlack import slack

from . import common_options

_attach_help = (
    "A key value pair, where the key is the title of the attachment ('' "
    + "if not needed) and the value is the attachment body. Can be provided multiple "
    + "times, such as in -a key1 val1 -a key2 val2."
)


@click.command()
@click.argument("message", nargs=1, default="", type=str)
@click.option(
    "-a", "--attach", "attach", nargs=2, multiple=True, type=str, help=_attach_help
)
@click.option(
    "-u",
    "--upload",
    "upload",
    type=click.Path(exists=True),
    help="Option to upload a file.",
)
@common_options
def main(oauth_api_token, channel, message, attach, upload):
    """Send a message and/or attachments from the command line.

    Attachments can be provided as key-value pairs
    
    $ shlack message 'Look!' -a greet 'Hi!' -a date $(date) -c ... -t ...

    """
    # define messenger
    slacker = slack.slacker_factory(api_key=oauth_api_token)

    # format message elements
    if not attach:
        attachments = None
    else:
        attachments = [slack.attachment_formatter(dict(attach))]

    if message == "":
        message = None

    if upload is not None:

        if attachments is None:
            attachments = []

        permalink = slack.upload_file_get_permalink(slacker, file_=upload)
        attachments.append(slack.attachment_formatter({upload: permalink}))

    # exit if nothing to do
    if (None, None, None) == (message, attachments, upload):
        sys.exit(0)

    slacker.chat.post_message(
        channel, text=message, attachments=attachments, unfurl_links=True
    )


if __name__ == "__main__":
    main()



"""Utilities for interacting with the slack API."""
import copy
import os
import subprocess
import time

from slacker import Slacker

MAX_TEXT_LENGTH = 750


class SlackError(Exception):
    """Custom error for slack issues."""

    pass


def slacker_factory(api_key=None):
    """Return a slacker object, getting the API key from the user or env."""
    if api_key is None:
        api_key = os.getenv("SLACK_OAUTH_API_TOKEN")
        if api_key is None:
            raise EnvironmentError("No Oauth API key provided!")

    return Slacker(api_key)


def upload_file_get_permalink(slacker, raise_error=False, *args, **kwargs):
    """Upload a file and return the permalink.

    Option to raise SlackError is there is a problem, or silently return None.

    This simply handles the return value of slacker.files.upload. The signature for 
    slacker.files.upload is:

    file_=None,
    content=None,
    filetype=None,
    filename=None,
    title=None,
    initial_comment=None,
    channels=None,
    thread_ts=None
    """
    out = slacker.files.upload(*args, **kwargs).body

    if raise_error and not out["ok"]:
        raise SlackError("Unable to upload file.")

    return out.get("file", {}).get("permalink")


def attachment_formatter(attachment, color="#DCDEE0"):
    """Format a dict or string to become a slack attachment.

    If attachment is a dict, it will gave one item per key, titled by the name of the 
    key. If a string, there will be one item with an empty title.
    
    Basically copied verbatim from jarjar. Thx jeff :-)
    """
    result = dict(fallback="", color=color, fields=[], ts=time.time())
    attachment_dict = attachment if hasattr(attachment, "keys") else {"": attachment}
    for key in attachment_dict:

        if isinstance(attachment_dict[key], str):
            outval = attachment_dict[key]
        else:
            try:
                outval = str(attachment_dict[key])
            except UnicodeEncodeError:
                outval = unicode(attachment_dict[key])

        result["fields"].append(dict(title=key, value=outval, short=len(outval) < 20))
    return result


def attach_text_data(slacker, text_data, name, data_format, color):
    """Create an attachment for a blob of text data.

    Creates a monospace formatted attachment OR text file depending on user input
    and data length.

    - Attach a permalink to a file if asked or on auto and the data are too big.
    - Attach plain text if asked or if on auto and the data are small.
    - If attaching text explicitly and data are too big, text will be clipped.
    - Return None otherwise.
    """
    if data_format not in ("auto", "text", "file"):
        return

    is_long = len(text_data) > MAX_TEXT_LENGTH

    # if asked for a file, or if auto and the text are long, return a permalink to
    # the file
    if data_format == "file" or (data_format == "auto" and is_long):
        permalink = upload_file_get_permalink(
            slacker, content=text_data, filetype="text", filename=name, title=name
        )
        content = permalink or "error uploading file."
        title = "%s upload" % name
        return attachment_formatter({title: content}, color=color)

    # attach if the user asked or auto and the text are short
    if data_format == "text" or (data_format == "auto" and not is_long):
        content = text_data[:MAX_TEXT_LENGTH] + "\n[ ... ]" if is_long else text_data
        return attachment_formatter({"": "```%s```" % (content)}, color=color)



"""Setup the package."""

import os

from setuptools import find_packages, setup

# use readme as long description
README_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "readme.md")

with open(README_PATH) as file_:
    LONG_DESCRIPTION = file_.read()

setup(
    name="shlack",
    version="0.0.1",
    description="Yet another slack command line interface.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Nolan Conaway",
    author_email="nolanbconaway@gmail.com",
    url="https://github.com/nolanbconaway/shlack",
    classifiers=[
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords=["slack", "cli"],
    license="MIT",
    packages=["shlack", "shlack.cli"],
    install_requires=["slacker", "click"],
    entry_points={"console_scripts": ["shlack = shlack.cli.__main__:cli"]},
)



"""Tests for the slack module."""
import pytest

from shlack import slack


@pytest.mark.parametrize(
    "attachment,formatted",
    [
        (
            "non-dict is made into one",
            dict(
                fallback="",
                color="",
                fields=[dict(title="", value="non-dict is made into one", short=False)],
                ts=0,
            ),
        ),
        (
            {1: "int key left alone"},
            dict(
                fallback="",
                color="",
                fields=[dict(title=1, value="int key left alone", short=True)],
                ts=0,
            ),
        ),
        (
            {"int value is stringed": 1},
            dict(
                fallback="",
                color="",
                fields=[dict(title="int value is stringed", value="1", short=True)],
                ts=0,
            ),
        ),
        (
            {"int value is stringed": 1},
            dict(
                fallback="",
                color="",
                fields=[dict(title="int value is stringed", value="1", short=True)],
                ts=0,
            ),
        ),
        (
            {1: 1, 2: 2},
            dict(
                fallback="",
                color="",
                fields=[
                    dict(title=1, value="1", short=True),
                    dict(title=2, value="2", short=True),
                ],
                ts=0,
            ),
        ),
    ],
)
def test_attachment_formatter(monkeypatch, attachment, formatted):
    """Test the attachment formatter."""
    monkeypatch.setattr("time.time", lambda: 0)
    assert slack.attachment_formatter(attachment, color="") == formatted


def test_slacker_factor_no_key_error(monkeypatch):
    """Test that an error is raise if there is no API key."""
    monkeypatch.delenv("SLACK_OAUTH_API_TOKEN", raising=False)
    with pytest.raises(EnvironmentError):
        slack.slacker_factory()


def test_upload_file_get_permalink(monkeypatch):
    """Test the file uploader behaves."""
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")

    class Object(object):
        pass

    slacker = Object()
    setattr(slacker, "files", Object())
    resp = Object()
    setattr(resp, "body", dict(ok=True, file=dict(permalink="URL")))
    setattr(slacker.files, "upload", lambda *a, **kw: resp)

    assert slack.upload_file_get_permalink(slacker) == "URL"


def test_upload_file_get_permalink_error(monkeypatch):
    """Test the file uploader behaves."""
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")

    class Object(object):
        pass

    slacker = Object()
    setattr(slacker, "files", Object())
    resp = Object()
    setattr(resp, "body", dict(ok=False))
    setattr(slacker.files, "upload", lambda *a, **kw: resp)

    print(resp.body.get("file", {}).get("permalink"))
    with pytest.raises(slack.SlackError):
        slack.upload_file_get_permalink(slacker, raise_error=True)


def test_attach_text_data_file(monkeypatch):
    """Test the task output attchment logic when a file is expected."""
    monkeypatch.setattr(
        "shlack.slack.upload_file_get_permalink", lambda *a, **kw: "URL"
    )
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")
    monkeypatch.setenv("SLACK_CHANNEL", "ALSO_FAKE")
    field = slack.attach_text_data(
        slacker="fake", text_data="data", name="name", data_format="file", color=""
    )["fields"][0]
    assert "name" in field["title"]
    assert "URL" in field["value"]

    # add auto format test
    field = slack.attach_text_data(
        slacker="fake", text_data="1" * 1001, name="name", data_format="auto", color=""
    )["fields"][0]
    assert "name" in field["title"]
    assert "URL" in field["value"]


def test_attach_text_data_text(monkeypatch):
    """Test the task output attchment logic when text is expected."""
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")
    monkeypatch.setenv("SLACK_CHANNEL", "ALSO_FAKE")
    field = slack.attach_text_data(
        slacker="fake", text_data="data", name="name", data_format="text", color=""
    )["fields"][0]
    assert not field["title"]
    assert "data" in field["value"]

    # auto format test
    field = slack.attach_text_data(
        slacker="fake", text_data="data", name="name", data_format="auto", color=""
    )["fields"][0]
    assert not field["title"]
    assert "data" in field["value"]


def test_attach_text_data_none(monkeypatch):
    """Test the task output attchment logic when nothing is expected."""
    monkeypatch.setenv("SLACK_OAUTH_API_TOKEN", "FAKE")
    monkeypatch.setenv("SLACK_CHANNEL", "ALSO_FAKE")
    assert None is slack.attach_text_data(
        slacker="fake", text_data="data", name="name", data_format="none", color=""
    )



"""Init the shlack cli module."""
import click


def common_options(func):
    """Append common args to the command."""
    for option in reversed(
        [
            click.option(
                "-c",
                "--channel",
                "channel",
                type=str,
                envvar="SLACK_CHANNEL",
                help=(
                    "Channel to post within. Read from $SLACK_CHANNEL if not provided. "
                    "This can be a slack @username (though that is now deprecated), #channel or "
                    "user/channel ID."
                ),
            ),
            click.option(
                "-t",
                "--api-token",
                "oauth_api_token",
                default=None,
                type=str,
                envvar="SLACK_OAUTH_API_TOKEN",
                help="Oauth API Token. Read from $SLACK_OAUTH_API_TOKEN if not provided.",
            ),
        ]
    ):
        func = option(func)
    return func


@click.group()
def cli():
    """Send slack messages from the command line.
    
    This command line tool does nothing in itself, call one of the below described 
    subcommands to get started!
    """
    # Do nothing. All function is in the subcommands.
    pass
